#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP
from holidays.constants import OCT, NOV, DEC, SUN
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class Spain(HolidayBase):
    country = "ES"
    subdivisions = [
        "AN",
        "AR",
        "AS",
        "CB",
        "CE",
        "CL",
        "CM",
        "CN",
        "CT",
        "EX",
        "GA",
        "IB",
        "MC",
        "MD",
        "ML",
        "NC",
        "PV",
        "RI",
        "VC",
    ]

    def _is_observed(self, date_holiday, name_holiday):
        if self.observed and date_holiday.weekday() == SUN:
            self[date_holiday + rd(days=+1)] = name_holiday + " (Trasladado)"
        else:
            self[date_holiday] = name_holiday

    def _populate(self, year):
        super()._populate(year)

        self._is_observed(date(year, JAN, 1), "Año nuevo")
        self._is_observed(date(year, JAN, 6), "Epifanía del Señor")

        easter_date = easter(year)

        if (
            year < 2015
            and self.subdiv
            and self.subdiv
            in {
                "AR",
                "CL",
                "CM",
                "EX",
                "GA",
                "MC",
                "MD",
                "ML",
                "NC",
                "PV",
                "VC",
            }
        ):
            self._is_observed(date(year, MAR, 19), "San José")
        elif (
            year == 2015
            and self.subdiv
            and self.subdiv in {"CM", "MC", "MD", "ML", "NC", "PV", "VC"}
        ):
            self._is_observed(date(year, MAR, 19), "San José")
        elif (
            year == 2016
            and self.subdiv
            and self.subdiv in {"MC", "ML", "PV", "VC"}
        ):
            self._is_observed(date(year, MAR, 19), "San José")
        elif year == 2017 and self.subdiv in {"PV"}:
            self._is_observed(date(year, MAR, 19), "San José")
        elif (
            2018 <= year <= 2019
            and self.subdiv
            and self.subdiv in {"GA", "MC", "NC", "PV", "VC"}
        ):
            self._is_observed(date(year, MAR, 19), "San José")
        elif (
            2020 <= year <= 2021
            and self.subdiv
            and self.subdiv in {"CM", "GA", "MC", "NC", "PV", "VC"}
        ):
            self._is_observed(date(year, MAR, 19), "San José")
        elif year >= 2022 and self.subdiv and self.subdiv == "VC":
            self._is_observed(date(year, MAR, 19), "San José")
        if year != 2022 and self.subdiv not in {"CT", "VC"}:
            self[easter_date + rd(days=-3)] = "Jueves Santo"
        elif year == 2022 and self.subdiv and self.subdiv not in {"CT"}:
            self[easter_date + rd(days=-3)] = "Jueves Santo"
        self[easter_date + rd(days=-2)] = "Viernes Santo"
        if (
            2022 == year
            and self.subdiv
            and self.subdiv in {"CT", "IB", "NC", "PV", "RI", "VC"}
        ):
            self[easter_date + rd(days=+1)] = "Lunes de Pascua"
        elif 2022 > year and self.subdiv in {
            "CM",
            "CT",
            "IB",
            "NC",
            "PV",
            "VC",
        }:
            self[easter(year) + rd(days=+1)] = "Lunes de Pascua"

        if 2022 != year:
            self._is_observed(date(year, MAY, 1), "Día del Trabajador")
        elif 2022 == year and self.subdiv in {
            "AN",
            "AR",
            "AS",
            "CL",
            "EX",
            "MC",
        }:
            self._is_observed(date(year, MAY, 1), "Día del Trabajador")
        if self.subdiv in {"CT", "GA", "VC"}:
            self._is_observed(date(year, JUN, 24), "San Juan")
        self._is_observed(date(year, AUG, 15), "Asunción de la Virgen")
        self._is_observed(date(year, OCT, 12), "Día de la Hispanidad")
        self._is_observed(date(year, NOV, 1), "Todos los Santos")
        self._is_observed(
            date(year, DEC, 6), "Día de la Constitución Española"
        )
        self._is_observed(date(year, DEC, 8), "La Inmaculada Concepción")
        if year != 2022:
            self._is_observed(date(year, DEC, 25), "Navidad")
        elif year == 2022 and self.subdiv not in {"CE", "GA", "PV", "VC"}:
            self._is_observed(date(year, DEC, 26), "Navidad (Trasladado)")
        if self.subdiv in {"CT", "IB"}:
            self._is_observed(date(year, DEC, 26), "San Esteban")

        # Provinces festive day
        if self.subdiv == "AN":
            self._is_observed(date(year, FEB, 28), "Día de Andalucia")
        elif self.subdiv == "AR":
            self._is_observed(date(year, APR, 23), "Día de San Jorge")
        elif self.subdiv == "AS":
            self._is_observed(date(year, SEP, 8), "Día de Asturias")
        elif self.subdiv == "CB":
            self._is_observed(
                date(year, JUL, 28),
                "Día de las Instituciones de Cantabria",
            )
            self._is_observed(date(year, SEP, 15), "Día de la Bien Aparecida")
        elif self.subdiv == "CE":
            self._is_observed(date(year, AUG, 5), "Nuestra Señora de África")
            self._is_observed(
                date(year, SEP, 2), "Día de la Ciudad Autónoma de Ceuta"
            )
            if year >= 2022:
                self._is_observed(_islamic_to_gre(year, 12, 10)[0], "Eid Adha")
        elif self.subdiv == "CM":
            if year >= 2022:
                self._is_observed(date(year, JUN, 16), "Corpus Christi")
            self._is_observed(date(year, MAY, 31), "Día de Castilla La Mancha")
        elif self.subdiv == "CN":
            self._is_observed(date(year, MAY, 30), "Día de Canarias")
        elif self.subdiv == "CL":
            self._is_observed(date(year, APR, 23), "Día de Castilla y Leon")
        elif self.subdiv == "CT":
            if year >= 2022:
                self._is_observed(
                    date(year, JUN, 6), "Día de la Pascua Granada"
                )
            self._is_observed(date(year, SEP, 11), "Día Nacional de Catalunya")
        elif self.subdiv == "EX":
            self._is_observed(date(year, SEP, 8), "Día de Extremadura")
        elif self.subdiv == "GA":
            if year >= 2022:
                self._is_observed(
                    date(year, MAY, 17), "Día de las letras Gallegas"
                )
            self._is_observed(date(year, JUL, 25), "Día Nacional de Galicia")
        elif self.subdiv == "IB":
            self._is_observed(date(year, MAR, 1), "Día de las Islas Baleares")
        elif self.subdiv == "MD":
            self._is_observed(date(year, MAY, 2), "Día de Comunidad de Madrid")
            if year >= 2022:
                self._is_observed(
                    date(year, JUL, 25), "Día de Santiago Apóstol"
                )
        elif self.subdiv == "MC":
            self._is_observed(date(year, JUN, 9), "Día de la Región de Murcia")
        elif self.subdiv == "ML":
            self._is_observed(date(year, SEP, 8), "Vírgen de la victoria")
            self._is_observed(date(year, SEP, 17), "Día de Melilla")
            if year == 2022:
                self._is_observed(
                    _islamic_to_gre(year, 10, 1)[0] + rd(days=+1),
                    "Aid Al-Fitr",
                )
                self._is_observed(
                    _islamic_to_gre(year, 12, 10)[0] + rd(days=+2),
                    "Aid Al-Adha",
                )
            else:
                self._is_observed(
                    _islamic_to_gre(year, 10, 1)[0], "Aid Al-Fitr"
                )
                self._is_observed(
                    _islamic_to_gre(year, 12, 10)[0], "Aid Al-Adha"
                )
        elif self.subdiv == "NC":
            if year >= 2022:
                self._is_observed(
                    date(year, JUL, 25), "Día de Santiago Apóstol"
                )
            self._is_observed(date(year, SEP, 27), "Día de Navarra")
        elif self.subdiv == "PV":
            if year >= 2022:
                self._is_observed(
                    date(year, JUL, 25), "Día de Santiago Apóstol"
                )
                self._is_observed(date(year, SEP, 6), "Día de Elcano")
            if 2011 <= year <= 2013:
                self._is_observed(date(year, OCT, 25), "Día del País Vasco")
        elif self.subdiv == "RI":
            self._is_observed(date(year, JUN, 9), "Día de La Rioja")
        elif self.subdiv == "VC":
            if year <= 2021:
                self._is_observed(
                    date(year, OCT, 9), "Día de la Comunidad Valenciana"
                )


class ES(Spain):
    pass


class ESP(Spain):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)


import warnings
from datetime import date
from typing import Tuple

from dateutil.relativedelta import relativedelta as rd

# Installation: pip install korean_lunar_calendar
# URL: https://github.com/usingsky/korean_lunar_calendar_py/
from korean_lunar_calendar import KoreanLunarCalendar

from holidays.constants import JAN, MAR, APR, MAY, JUN, JUL, AUG, OCT, DEC
from holidays.constants import SAT, SUN
from holidays.holiday_base import HolidayBase


class SouthKorea(HolidayBase):
    """
    1. https://publicholidays.co.kr/ko/2020-dates/
    2. https://en.wikipedia.org/wiki/Public_holidays_in_South_Korea
    3. https://www.law.go.kr/%EB%B2%95%EB%A0%B9/%EA%B4%80%EA%B3%B5%EC%84%9C%EC%9D%98%20%EA%B3%B5%ED%9C%B4%EC%9D%BC%EC%97%90%20%EA%B4%80%ED%95%9C%20%EA%B7%9C%EC%A0%95  # noqa

    According to (3), the alt holidays in Korea are as follows:
    The alt holiday means next first non holiday after the holiday.
    Independence movement day, Liberation day, National Foundation Day,
    Hangul Day, Children's Day have alt holiday if they fell on saturday or sunday.
    Lunar New Year's Day, Korean Mid Autumn Day have alt holiday if they fell
    on only sunday.

    """

    country = "KR"
    special_holidays = {
        # Just for year 2020 - since 2020.08.15 is Sat, the government
        # decided to make 2020.08.17 holiday, yay
        2020: ((AUG, 17, "Alternative public holiday"),)
    }

    def __init__(self, **kwargs):
        self.korean_cal = KoreanLunarCalendar()
        HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        alt_holiday = "Alternative holiday of "

        # New Year's Day
        name = "New Year's Day"
        self[date(year, JAN, 1)] = name

        # Lunar New Year
        name = "Lunar New Year's Day"
        preceding_day_lunar = "The day preceding of " + name
        second_day_lunar = "The second day of " + name

        dt = self.get_solar_date(year, 1, 1)
        new_year_date = date(dt.year, dt.month, dt.day)

        self[new_year_date + rd(days=-1)] = preceding_day_lunar
        self[new_year_date] = name
        self[new_year_date + rd(days=+1)] = second_day_lunar

        if self.observed and year >= 2015:
            for cur_rd, cur_name in [
                (-1, preceding_day_lunar),
                (0, name),
                (+1, second_day_lunar),
            ]:
                target_date = new_year_date + rd(days=cur_rd)
                is_alt, alt_date = self.get_next_first_non_holiday(
                    cur_name, target_date
                )
                if is_alt:
                    self[alt_date] = alt_holiday + name

        # Independence Movement Day
        name = "Independence Movement Day"
        independence_date = date(year, MAR, 1)

        self[independence_date] = name

        if self.observed and year >= 2021:
            is_alt, alt_date = self.get_next_first_non_holiday(
                name, independence_date, include_sat=True
            )
            if is_alt:
                self[alt_date] = alt_holiday + name

        # Tree Planting Day
        # removed from holiday since 2006
        name = "Tree Planting Day"
        planting_date = date(year, APR, 5)
        if self.observed and 1949 <= year <= 2005 and year != 1960:
            self[planting_date] = name

        # Birthday of the Buddha
        name = "Birthday of the Buddha"
        dt = self.get_solar_date(year, 4, 8)
        buddha_date = date(dt.year, dt.month, dt.day)
        self[buddha_date] = name

        # Children's Day
        name = "Children's Day"
        childrens_date = date(year, MAY, 5)
        if year >= 1975:
            self[childrens_date] = name
            if self.observed and year >= 2015:
                is_alt, alt_date = self.get_next_first_non_holiday(
                    name, childrens_date, include_sat=True
                )
                if is_alt:
                    self[alt_date] = alt_holiday + name
        else:
            # no children's day before 1975
            pass

        # Labour Day
        name = "Labour Day"
        if year >= 1994:
            labour_date = date(year, MAY, 1)
        else:
            labour_date = date(year, MAR, 10)
        self[labour_date] = name

        # Memorial Day
        name = "Memorial Day"
        memorial_date = date(year, JUN, 6)
        self[memorial_date] = name

        # Constitution Day
        # removed from holiday since 2008
        name = "Constitution Day"
        constitution_date = date(year, JUL, 17)
        if self.observed and 1948 <= year <= 2007:
            self[constitution_date] = name

        # Liberation Day
        name = "Liberation Day"
        libration_date = date(year, AUG, 15)
        if year >= 1945:
            self[libration_date] = name
            if self.observed and year >= 2021:
                is_alt, alt_date = self.get_next_first_non_holiday(
                    name, libration_date, include_sat=True
                )
                if is_alt:
                    self[alt_date] = alt_holiday + name
        else:
            pass

        # Korean Mid Autumn Day
        name = "Chuseok"
        preceding_day_chuseok = "The day preceding of " + name
        second_day_chuseok = "The second day of " + name

        dt = self.get_solar_date(year, 8, 15)
        chuseok_date = date(dt.year, dt.month, dt.day)

        self[chuseok_date + rd(days=-1)] = preceding_day_chuseok
        self[chuseok_date] = name
        self[chuseok_date + rd(days=+1)] = second_day_chuseok

        if self.observed and year >= 2014:
            for cur_rd, cur_name in [
                (-1, preceding_day_chuseok),
                (0, name),
                (+1, second_day_chuseok),
            ]:
                target_date = chuseok_date + rd(days=cur_rd)
                is_alt, alt_date = self.get_next_first_non_holiday(
                    cur_name, target_date
                )
                if is_alt:
                    self[alt_date] = alt_holiday + name

        # National Foundation Day
        name = "National Foundation Day"
        foundation_date = date(year, OCT, 3)
        self[foundation_date] = name

        if self.observed and year >= 2021:
            is_alt, alt_date = self.get_next_first_non_holiday(
                name, foundation_date, include_sat=True
            )
            if is_alt:
                self[alt_date] = alt_holiday + name

        # Hangul Day
        name = "Hangeul Day"
        if year <= 1990 or year >= 2013:
            hangeul_date = date(year, OCT, 9)
            self[hangeul_date] = name

            if self.observed and year >= 2021:
                is_alt, alt_date = self.get_next_first_non_holiday(
                    name, hangeul_date, include_sat=True
                )
                if is_alt:
                    self[alt_date] = alt_holiday + name

        # Christmas Day
        name = "Christmas Day"
        christmas_date = date(year, DEC, 25)
        self[christmas_date] = name

    # convert lunar calendar date to solar
    def get_solar_date(self, year: int, month: int, day: int) -> date:
        """Return the Gregorian calendar date of a Korean lunar calendar date.

        :param year:
           The Korean lunar year(년).

        :param month:
           The Korean lunar month(월).

        :param day:
           The Korean lunar day(일).

        :return:
           The Korean Gregorian date.
        """
        self.korean_cal.setLunarDate(year, month, day, False)
        return date(
            self.korean_cal.solarYear,
            self.korean_cal.solarMonth,
            self.korean_cal.solarDay,
        )

    def get_next_first_non_holiday(
        self, name: str, cur: date, include_sat: bool = False
    ) -> Tuple[bool, date]:
        """Returns the first day from the date provided that's not already a
        holiday of a different name nor a weekend.

        :param name:
           The name of the holiday.

        :param cur:
           The current date of the holiday.

        :param include_sat:
           Whether Saturday is to be considered a weekend in addition to
           Sunday.

        :return:
           A tuple consisting of a flag set to whether the date is different
           and the date itself.

        """

        start_value = cur
        target_weekday = [SUN]
        if include_sat:
            target_weekday.append(SAT)
        check_1 = cur.weekday() in target_weekday  # Exclude weekends
        check_2 = (
            cur in self and name != self[cur]
        )  # Exclude if already a holiday
        while check_1 or check_2:
            cur = cur + rd(days=1)
            check_1 = cur.weekday() in target_weekday
            check_2 = cur in self and name != self[cur]

        return start_value != cur, cur


class Korea(SouthKorea):
    def __init__(self, *args, **kwargs) -> None:
        warnings.warn(
            "Korea is deprecated, use SouthKorea instead.",
            DeprecationWarning,
        )

        super().__init__(*args, **kwargs)


class KR(SouthKorea):
    pass


class KOR(SouthKorea):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import MO
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, APR, MAY, AUG, DEC, SUN
from holidays.holiday_base import HolidayBase


class Zimbabwe(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Robert_Mugabe
    https://en.wikipedia.org/wiki/Public_holidays_in_Zimbabwe
    """

    country = "ZW"

    def _populate(self, year):
        if year <= 1987:
            return
        super()._populate(year)

        self[date(year, JAN, 1)] = "New Year's Day"

        # https://en.wikipedia.org/wiki/Robert_Gabriel_Mugabe_National_Youth_Day
        if year >= 2018:
            self[
                date(year, FEB, 21)
            ] = "Robert Gabriel Mugabe National Youth Day"

        easter_date = easter(year)
        self[easter_date + rd(days=-2)] = "Good Friday"
        self[easter_date + rd(days=-1)] = "Easter Saturday"
        self[easter_date + rd(days=+1)] = "Easter Monday"

        self[date(year, APR, 18)] = "Independence Day"

        self[date(year, MAY, 1)] = "Workers' Day"
        self[date(year, MAY, 25)] = "Africa Day"

        # 2nd Monday of August
        zimbabwe_heroes_day = date(year, AUG, 1) + rd(weekday=MO(+2))
        self[zimbabwe_heroes_day] = "Zimbabwe Heroes' Day"

        # Tuesday after 2nd Monday of August
        defence_forces_day = zimbabwe_heroes_day + rd(days=1)
        self[defence_forces_day] = "Defense Forces Day"

        self[date(year, DEC, 22)] = "Unity Day"
        self[date(year, DEC, 25)] = "Christmas Day"
        self[date(year, DEC, 26)] = "Boxing Day"

        if self.observed:
            for k, v in list(self.items()):
                if k.weekday() == SUN and k.year == year:
                    dt = k + rd(days=+1)
                    while self.get(dt):
                        dt += rd(days=+1)
                    self[dt] = v + " (Observed)"


class ZW(Zimbabwe):
    pass


class ZWE(Zimbabwe):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, APR, MAY, OCT
from holidays.holiday_base import HolidayBase
from holidays.utils import _ChineseLuniSolar


class China(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_China
    """

    country = "CN"

    def __init__(self, **kwargs):
        self.cnls = _ChineseLuniSolar()
        HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        if year <= 1949:
            return

        self[date(year, JAN, 1)] = "New Year's Day"

        name = "Labour Day"
        dt = date(year, MAY, 1)
        self[dt] = name
        if 2000 <= year <= 2007:
            self[dt + rd(days=+1)] = name
            self[dt + rd(days=+2)] = name

        name = "Chinese New Year (Spring Festival)"
        dt = self.cnls.lunar_n_y_date(year)
        self[dt] = name
        self[dt + rd(days=+1)] = name
        if 2008 <= year <= 2013:
            self[dt + rd(days=-1)] = name
        else:
            self[dt + rd(days=+2)] = name

        name = "National Day"
        dt = date(year, OCT, 1)
        self[dt] = name
        self[dt + rd(days=+1)] = name
        if year >= 2000:
            self[dt + rd(days=+2)] = name

        if year >= 2008:
            self[date(year, APR, 5)] = "Tomb-Sweeping Day"
            self[self.cnls.lunar_to_gre(year, 5, 5)] = "Dragon Boat Festival"
            self[self.cnls.lunar_to_gre(year, 8, 15)] = "Mid-Autumn Festival"


class CN(China):
    pass


class CHN(China):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import MO, TU, TH, FR
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP
from holidays.constants import OCT, NOV, DEC, MON, WED, FRI, SAT, SUN
from holidays.holiday_base import HolidayBase


class UnitedStates(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_the_United_States
    """

    country = "US"
    subdivisions = [
        "AL",
        "AK",
        "AS",
        "AZ",
        "AR",
        "CA",
        "CO",
        "CT",
        "DE",
        "DC",
        "FL",
        "GA",
        "GU",
        "HI",
        "ID",
        "IL",
        "IN",
        "IA",
        "KS",
        "KY",
        "LA",
        "ME",
        "MD",
        "MH",
        "MA",
        "MI",
        "FM",
        "MN",
        "MS",
        "MO",
        "MT",
        "NE",
        "NV",
        "NH",
        "NJ",
        "NM",
        "NY",
        "NC",
        "ND",
        "MP",
        "OH",
        "OK",
        "OR",
        "PW",
        "PA",
        "PR",
        "RI",
        "SC",
        "SD",
        "TN",
        "TX",
        "UT",
        "VT",
        "VA",
        "VI",
        "WA",
        "WV",
        "WI",
        "WY",
    ]

    def _add_with_observed(
        self, dt: date, name: str, before: bool = True, after: bool = True
    ) -> None:
        self[dt] = name
        if not self.observed:
            return
        if dt.weekday() == SAT and before:
            self[dt + rd(days=-1)] = f"{name} (Observed)"
        elif dt.weekday() == SUN and after:
            self[dt + rd(days=+1)] = f"{name} (Observed)"

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        if year >= 1871:
            name = "New Year's Day"
            self._add_with_observed(date(year, JAN, 1), name, before=False)
            # The following year's observed New Year's Day can be in this year
            # when it falls on a Friday (Jan 1st is a Saturday).
            if self.observed and date(year, DEC, 31).weekday() == FRI:
                self[date(year, DEC, 31)] = f"{name} (Observed)"

        # Epiphany
        if self.subdiv == "PR":
            self[date(year, JAN, 6)] = "Epiphany"

        # Three King's Day
        if self.subdiv == "VI":
            self[date(year, JAN, 6)] = "Three King's Day"

        # Lee Jackson Day
        if self.subdiv == "VA" and year <= 2020:
            name = "Lee Jackson Day"
            if year >= 2000:
                dt = date(year, JAN, 1) + rd(weekday=MO(+3)) + rd(days=-3)
                self[dt] = name
            elif year >= 1983:
                self[date(year, JAN, 1) + rd(weekday=MO(+3))] = name
            elif year >= 1889:
                self[date(year, JAN, 19)] = name

        # Inauguration Day
        if (
            self.subdiv in {"DC", "LA", "MD", "VA"}
            and year >= 1789
            and (year - 1789) % 4 == 0
        ):
            if year >= 1937:
                dt = date(year, JAN, 20)
            else:
                dt = date(year, MAR, 4)
            self._add_with_observed(dt, "Inauguration Day", before=False)

        # Martin Luther King Jr. Day
        if year >= 1986:
            name = "Martin Luther King Jr. Day"
            if self.subdiv == "AL":
                name = "Robert E. Lee/Martin Luther King Birthday"
            elif (self.subdiv == "MS") or (
                (self.subdiv == "AR") and (year <= 2017)
            ):
                name = (
                    "Dr. Martin Luther King Jr. "
                    "and Robert E. Lee's Birthdays"
                )
            elif self.subdiv in {"AZ", "NH"}:
                name = "Dr. Martin Luther King Jr./Civil Rights Day"
            elif self.subdiv == "GA" and year <= 2011:
                name = "Robert E. Lee's Birthday"
            elif self.subdiv == "ID" and year >= 2006:
                name = "Martin Luther King Jr. - Idaho Human Rights Day"
            self[date(year, JAN, 1) + rd(weekday=MO(+3))] = name

        # Lincoln's Birthday
        if (
            self.subdiv in {"CT", "IA", "IL", "NJ", "NY"} and year >= 1971
        ) or (self.subdiv == "CA" and 1971 <= year <= 2009):
            self._add_with_observed(date(year, FEB, 12), "Lincoln's Birthday")

        # Susan B. Anthony Day
        if (
            (self.subdiv == "CA" and year >= 2014)
            or (self.subdiv == "FL" and year >= 2011)
            or (self.subdiv == "NY" and year >= 2004)
            or (self.subdiv == "WI" and year >= 1976)
        ):
            self[date(year, FEB, 15)] = "Susan B. Anthony Day"

        # Washington's Birthday
        name = "Washington's Birthday"
        if self.subdiv == "AL":
            name = "George Washington/Thomas Jefferson Birthday"
        elif self.subdiv == "AR":
            name = "George Washington's Birthday and Daisy Gatson Bates Day"
        elif self.subdiv in {"PR", "VI"}:
            name = "Presidents' Day"
        if self.subdiv not in {"DE", "FL", "GA", "NM", "PR"}:
            if year >= 1971:
                self[date(year, FEB, 1) + rd(weekday=MO(+3))] = name
            elif year >= 1879:
                self[date(year, FEB, 22)] = name
        elif self.subdiv == "GA":
            if date(year, DEC, 24).weekday() != WED:
                self[date(year, DEC, 24)] = name
            else:
                self[date(year, DEC, 26)] = name
        elif self.subdiv in {"PR", "VI"}:
            self[date(year, FEB, 1) + rd(weekday=MO(+3))] = name

        easter_date = easter(year)
        # Mardi Gras
        if self.subdiv == "LA" and year >= 1857:
            self[easter_date + rd(days=-47)] = "Mardi Gras"

        # Guam Discovery Day
        if self.subdiv == "GU" and year >= 1970:
            self[date(year, MAR, 1) + rd(weekday=MO)] = "Guam Discovery Day"

        # Casimir Pulaski Day
        if self.subdiv == "IL" and year >= 1978:
            self[date(year, MAR, 1) + rd(weekday=MO)] = "Casimir Pulaski Day"

        # Texas Independence Day
        if self.subdiv == "TX" and year >= 1874:
            self[date(year, MAR, 2)] = "Texas Independence Day"

        # Town Meeting Day
        if self.subdiv == "VT" and year >= 1800:
            self[date(year, MAR, 1) + rd(weekday=TU)] = "Town Meeting Day"

        # Evacuation Day
        if self.subdiv == "MA" and year >= 1901:
            name = "Evacuation Day"
            dt = date(year, MAR, 17)
            self[dt] = name
            if self.observed and self._is_weekend(dt):
                self[dt + rd(weekday=MO)] = f"{name} (Observed)"

        # Emancipation Day
        if self.subdiv == "PR":
            self._add_with_observed(
                date(year, MAR, 22), "Emancipation Day", before=False
            )

        # Prince Jonah Kuhio Kalanianaole Day
        if self.subdiv == "HI" and year >= 1949:
            self._add_with_observed(
                date(year, MAR, 26), "Prince Jonah Kuhio Kalanianaole Day"
            )

        # Seward's Day
        if self.subdiv == "AK":
            name = "Seward's Day"
            if year >= 1955:
                self[date(year, MAR, 31) + rd(weekday=MO(-1))] = name
            elif year >= 1918:
                self[date(year, MAR, 30)] = name

        # César Chávez Day
        name = "César Chávez Day"
        dt = date(year, MAR, 31)
        if self.subdiv == "CA" and year >= 1995:
            self._add_with_observed(dt, name, before=False)
        elif self.subdiv == "TX" and year >= 2000:
            self[dt] = name

        # Transfer Day
        if self.subdiv == "VI":
            self[date(year, MAR, 31)] = "Transfer Day"

        # Emancipation Day
        if self.subdiv == "DC" and year >= 2005:
            self._add_with_observed(date(year, APR, 16), "Emancipation Day")

        # Patriots' Day
        if self.subdiv in {"MA", "ME"}:
            name = "Patriots' Day"
            if year >= 1969:
                self[date(year, APR, 1) + rd(weekday=MO(+3))] = name
            elif year >= 1894:
                self[date(year, APR, 19)] = name

        # Holy Thursday
        if self.subdiv == "VI":
            self[easter_date + rd(days=-3)] = "Holy Thursday"

        # Good Friday
        if self.subdiv in {
            "CT",
            "DE",
            "GU",
            "IN",
            "KY",
            "LA",
            "NC",
            "NJ",
            "PR",
            "TN",
            "TX",
            "VI",
        }:
            self[easter_date + rd(days=-2)] = "Good Friday"

        # Easter Monday
        if self.subdiv == "VI":
            self[easter_date + rd(days=+1)] = "Easter Monday"

        # Confederate Memorial Day
        name = "Confederate Memorial Day"
        if self.subdiv in {"AL", "GA", "MS", "SC"} and year >= 1866:
            if self.subdiv == "GA" and year >= 2016:
                name = "State Holiday"
            if self.subdiv == "GA" and year == 2020:
                self[date(year, APR, 10)] = name
            else:
                self[date(year, APR, 1) + rd(weekday=MO(+4))] = name
        elif self.subdiv == "TX" and year >= 1931:
            self[date(year, JAN, 19)] = name

        # San Jacinto Day
        if self.subdiv == "TX" and year >= 1875:
            self[date(year, APR, 21)] = "San Jacinto Day"

        # Arbor Day
        if self.subdiv == "NE":
            name = "Arbor Day"
            if year >= 1989:
                self[date(year, APR, 30) + rd(weekday=FR(-1))] = name
            elif year >= 1875:
                self[date(year, APR, 22)] = name

        # Primary Election Day
        if self.subdiv == "IN" and (
            (year >= 2006 and year % 2 == 0) or year >= 2015
        ):
            self[
                (date(year, MAY, 1) + rd(weekday=MO) + rd(days=+1))
            ] = "Primary Election Day"

        # Truman Day
        if self.subdiv == "MO" and year >= 1949:
            self._add_with_observed(date(year, MAY, 8), "Truman Day")

        # Memorial Day
        if year >= 1971:
            self[date(year, MAY, 31) + rd(weekday=MO(-1))] = "Memorial Day"
        elif year >= 1888:
            self[date(year, MAY, 30)] = "Memorial Day"

        # Juneteenth Day
        if year >= 2021:
            self._add_with_observed(
                date(year, JUN, 19), "Juneteenth National Independence Day"
            )

        # Jefferson Davis Birthday
        name = "Jefferson Davis Birthday"
        if self.subdiv == "AL" and year >= 1890:
            self[date(year, JUN, 1) + rd(weekday=MO)] = name

        # Kamehameha Day
        if self.subdiv == "HI" and year >= 1872:
            name = "Kamehameha Day"
            dt = date(year, JUN, 11)
            if year >= 2011:
                self._add_with_observed(dt, name)
            else:
                self[dt] = name

        # Emancipation Day In Texas
        if self.subdiv == "TX" and year >= 1980:
            self[date(year, JUN, 19)] = "Emancipation Day In Texas"

        # West Virginia Day
        if self.subdiv == "WV" and year >= 1927:
            self._add_with_observed(date(year, JUN, 20), "West Virginia Day")

        # Emancipation Day in US Virgin Islands
        if self.subdiv == "VI":
            self[date(year, JUL, 3)] = "Emancipation Day"

        # Independence Day
        if year >= 1871:
            self._add_with_observed(date(year, JUL, 4), "Independence Day")

        # Liberation Day (Guam)
        if self.subdiv == "GU" and year >= 1945:
            self[date(year, JUL, 21)] = "Liberation Day (Guam)"

        # Pioneer Day
        if self.subdiv == "UT" and year >= 1849:
            self._add_with_observed(date(year, JUL, 24), "Pioneer Day")

        # Constitution Day
        if self.subdiv == "PR":
            self._add_with_observed(
                date(year, JUL, 25), "Constitution Day", before=False
            )

        # Victory Day
        if self.subdiv == "RI" and year >= 1948:
            self[date(year, AUG, 1) + rd(weekday=MO(+2))] = "Victory Day"

        # Statehood Day (Hawaii)
        if self.subdiv == "HI" and year >= 1959:
            self[date(year, AUG, 1) + rd(weekday=FR(+3))] = "Statehood Day"

        # Bennington Battle Day
        if self.subdiv == "VT" and year >= 1778:
            self._add_with_observed(
                date(year, AUG, 16), "Bennington Battle Day"
            )

        # Lyndon Baines Johnson Day
        if self.subdiv == "TX" and year >= 1973:
            self[date(year, AUG, 27)] = "Lyndon Baines Johnson Day"

        # Labor Day
        if year >= 1894:
            self[date(year, SEP, 1) + rd(weekday=MO)] = "Labor Day"

        # Columbus Day
        if self.subdiv not in {"AK", "AR", "DE", "FL", "HI", "NV"}:
            if self.subdiv == "SD":
                name = "Native American Day"
            elif self.subdiv == "VI":
                name = "Columbus Day and Puerto Rico Friendship Day"
            else:
                name = "Columbus Day"
            if year >= 1970:
                self[date(year, OCT, 1) + rd(weekday=MO(+2))] = name
            elif year >= 1937:
                self[date(year, OCT, 12)] = name

        # Alaska Day
        if self.subdiv == "AK" and year >= 1867:
            self._add_with_observed(date(year, OCT, 18), "Alaska Day")

        # Nevada Day
        if self.subdiv == "NV" and year >= 1933:
            dt = date(year, OCT, 31)
            if year >= 2000:
                dt += rd(weekday=FR(-1))
            self._add_with_observed(dt, "Nevada Day")

        # Liberty Day
        if self.subdiv == "VI":
            self[date(year, NOV, 1)] = "Liberty Day"

        # Election Day
        if (
            self.subdiv
            in {"DE", "HI", "IL", "IN", "LA", "MT", "NH", "NJ", "NY", "WV"}
            and year >= 2008
            and year % 2 == 0
        ) or (self.subdiv in {"IN", "NY"} and year >= 2015):
            self[
                date(year, NOV, 1) + rd(weekday=MO) + rd(days=+1)
            ] = "Election Day"

        # All Souls' Day
        if self.subdiv == "GU":
            self[date(year, NOV, 2)] = "All Souls' Day"

        # Veterans Day
        if year >= 1954:
            name = "Veterans Day"
        else:
            name = "Armistice Day"
        if 1971 <= year <= 1977:
            self[date(year, OCT, 1) + rd(weekday=MO(+4))] = name
        elif year >= 1938:
            self._add_with_observed(date(year, NOV, 11), name)

        # Discovery Day
        if self.subdiv == "PR":
            self._add_with_observed(
                date(year, NOV, 19), "Discovery Day", before=False
            )

        # Thanksgiving
        if year >= 1871:
            self[date(year, NOV, 1) + rd(weekday=TH(+4))] = "Thanksgiving"

        # Day After Thanksgiving
        # Friday After Thanksgiving
        # Lincoln's Birthday
        # American Indian Heritage Day
        # Family Day
        # New Mexico Presidents' Day
        if (
            (
                self.subdiv in {"CA", "DE", "FL", "NC", "NH", "OK", "TX", "WV"}
                and year >= 1975
            )
            or (self.subdiv == "IN" and year >= 2010)
            or (self.subdiv == "MD" and year >= 2008)
            or self.subdiv in {"NM", "NV", "PA"}
        ):
            if self.subdiv in {"CA", "DE", "NC", "NH", "OK", "PA", "WV"}:
                name = "Day After Thanksgiving"
            if self.subdiv in {"FL", "TX"}:
                name = "Friday After Thanksgiving"
            if self.subdiv == "IN":
                name = "Lincoln's Birthday"
            if self.subdiv == "MD":
                name = "American Indian Heritage Day"
            if self.subdiv == "NV":
                name = "Family Day"
            if self.subdiv == "NM":
                name = "Presidents' Day"
            self[date(year, NOV, 1) + rd(weekday=TH(+4)) + rd(days=+1)] = name

        # Robert E. Lee's Birthday
        if self.subdiv == "GA" and year >= 1986:
            if year >= 2016:
                name = "State Holiday"
            else:
                name = "Robert E. Lee's Birthday"
            self[date(year, NOV, 29) + rd(weekday=FR(-1))] = name

        # Lady of Camarin Day
        if self.subdiv == "GU":
            self[date(year, DEC, 8)] = "Lady of Camarin Day"

        # Christmas Eve
        if (
            self.subdiv == "AS"
            or (self.subdiv in {"KS", "MI", "NC"} and year >= 2013)
            or (self.subdiv == "TX" and year >= 1981)
            or (self.subdiv == "WI" and year >= 2012)
        ):
            name = "Christmas Eve"
            dt = date(year, DEC, 24)
            self[dt] = name
            # If on Friday, observed on Thursday
            if self.observed and dt.weekday() == FRI:
                self[dt + rd(days=-1)] = f"{name} (Observed)"
            # If on Saturday or Sunday, observed on Friday
            elif self.observed and self._is_weekend(dt):
                self[dt + rd(weekday=FR(-1))] = f"{name} (Observed)"

        # Christmas Day
        if year >= 1871:
            self._add_with_observed(date(year, DEC, 25), "Christmas Day")

        # Day After Christmas
        dt = date(year, DEC, 26)
        if self.subdiv == "NC" and year >= 2013:
            name = "Day After Christmas"
            self[dt] = name
            # If on Saturday or Sunday, observed on Monday
            if self.observed and self._is_weekend(dt):
                self[dt + rd(weekday=MO)] = f"{name} (Observed)"
            # If on Monday, observed on Tuesday
            elif self.observed and dt.weekday() == MON:
                self[dt + rd(days=+1)] = f"{name} (Observed)"
        elif self.subdiv == "TX" and year >= 1981:
            self[dt] = "Day After Christmas"
        elif self.subdiv == "VI":
            self[dt] = "Christmas Second Day"

        # New Year's Eve
        if (self.subdiv in {"KY", "MI"} and year >= 2013) or (
            self.subdiv == "WI" and year >= 2012
        ):
            self._add_with_observed(
                date(year, DEC, 31), "New Year's Eve", before=True, after=False
            )


class US(UnitedStates):
    pass


class USA(UnitedStates):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import EASTER_ORTHODOX, easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAY, SEP, AUG, OCT, DEC
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class NorthMacedonia(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_North_Macedonia
    """

    country = "MK"

    def _populate(self, year):
        super()._populate(year)

        self[date(year, JAN, 1)] = "New Year's Day"

        self[date(year, JAN, 7)] = "Christmas Day (Orthodox)"
        easter_date = easter(year, method=EASTER_ORTHODOX)
        self[easter_date + rd(days=+1)] = "Easter Monday(Orthodox)"
        self[date(year, MAY, 1)] = "Labour Day"
        self[date(year, MAY, 24)] = "Saints Cyril and Methodius Day"
        self[date(year, AUG, 2)] = "Republic Day"
        self[date(year, SEP, 8)] = "Independence Day"
        self[date(year, OCT, 11)] = "Day of Macedonian Uprising in 1941"
        self[
            date(year, OCT, 23)
        ] = "Day of the Macedonian Revolutionary Struggle"
        self[date(year, DEC, 8)] = "Saint Clement of Ohrid Day"
        for date_obs in _islamic_to_gre(year, 10, 1):
            self[date_obs] = "Eid al-Fitr"


class MK(NorthMacedonia):
    pass


class MKD(NorthMacedonia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import DEC, JAN, MAY
from holidays.holiday_base import HolidayBase


class EuropeanCentralBank(HolidayBase):
    # https://en.wikipedia.org/wiki/TARGET2
    # http://www.ecb.europa.eu/press/pr/date/2000/html/pr001214_4.en.html

    market = "ECB"

    def __init__(self, **kwargs):
        HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        self[date(year, JAN, 1)] = "New Year's Day"
        e = easter(year)
        self[e - rd(days=2)] = "Good Friday"
        self[e + rd(days=1)] = "Easter Monday"
        self[date(year, MAY, 1)] = "1 May (Labour Day)"
        self[date(year, DEC, 25)] = "Christmas Day"
        self[date(year, DEC, 26)] = "26 December"


class ECB(EuropeanCentralBank):
    pass


class TAR(EuropeanCentralBank):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import MO
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, APR, MAY, JUL, AUG, SEP, OCT, DEC, SUN
from holidays.holiday_base import HolidayBase


class Zambia(HolidayBase):
    """
    https://www.officeholidays.com/countries/zambia/
    https://www.timeanddate.com/holidays/zambia/
    https://en.wikipedia.org/wiki/Public_holidays_in_Zambia
    https://www.parliament.gov.zm/sites/default/files/documents/acts/Public%20Holidays%20Act.pdf
    """

    country = "ZM"
    special_holidays = {
        2016: (
            (AUG, 11, "General elections and referendum"),
            (
                SEP,
                13,
                "Inauguration ceremony of President-elect "
                "and Vice President-elect",
            ),
        ),
        2018: (
            (MAR, 9, "Public holiday"),
            (JUL, 26, "Lusaka mayoral and other local government elections"),
        ),
        2021: (
            (JUL, 2, "Memorial service for Kenneth Kaunda"),
            (JUL, 7, "Funeral of Kenneth Kaunda"),
            (AUG, 12, "General elections"),
            (AUG, 13, "Counting in general elections"),
            (AUG, 24, "Presidential inauguration"),
        ),
        2022: ((MAR, 18, "Funeral of Rupiah Banda"),),
    }

    def _populate(self, year):

        # Observed since 1965
        if year <= 1964:
            return

        super()._populate(year)

        self[date(year, JAN, 1)] = "New Year's Day"

        if year >= 1991:
            self[date(year, MAR, 8)] = "International Women's Day"

        self[date(year, MAR, 12)] = "Youth Day"

        easter_date = easter(year)
        self[easter_date + rd(days=-2)] = "Good Friday"
        self[easter_date + rd(days=-1)] = "Holy Saturday"
        self[easter_date + rd(days=+1)] = "Easter Monday"

        if year >= 2022:
            self[date(year, APR, 28)] = "Kenneth Kaunda Day"

        self[date(year, MAY, 1)] = "Labour Day"
        self[date(year, MAY, 25)] = "Africa Freedom Day"

        # 1st Monday of July = "Heroes' Day"
        dt = date(year, JUL, 1) + rd(weekday=MO)
        self[dt] = "Heroes' Day"
        self[dt + rd(days=1)] = "Unity Day"

        # 1st Monday of Aug = "Farmers' Day"
        dt = date(year, AUG, 1) + rd(weekday=MO)
        self[dt] = "Farmers' Day"

        if year >= 2015:
            self[date(year, OCT, 18)] = "National Prayer Day"

        self[date(year, OCT, 24)] = "Independence Day"
        self[date(year, DEC, 25)] = "Christmas Day"

        # whenever a public holiday falls on a Sunday,
        # it rolls over to the following Monday
        if self.observed:
            for k, v in list(self.items()):
                if k.year == year and k.weekday() == SUN:
                    self[k + rd(days=1)] = v + " (Observed)"


class ZM(Zambia):
    pass


class ZMB(Zambia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from typing import Dict, Iterable, Optional, Tuple, Union

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAR, APR, MAY, JUN, JUL, SEP, AUG
from holidays.constants import OCT, NOV, DEC, SUN
from holidays.holiday_base import HolidayBase
from holidays.utils import _ChineseLuniSolar, _islamic_to_gre


class Singapore(HolidayBase):
    country = "SG"
    special_holidays = {
        2001: ((NOV, 3, "Polling Day"),),
        2006: ((MAY, 6, "Polling Day"),),
        2011: ((MAY, 7, "Polling Day"),),
        2015: (
            # SG50 Public holiday
            # Announced on 14 March 2015
            # https://www.mom.gov.sg/newsroom/press-releases/2015/sg50-public-holiday-on-7-august-2015
            (AUG, 7, "SG50 Public Holiday"),
            (SEP, 11, "Polling Day"),
        ),
        2020: ((JUL, 10, "Polling Day"),),
    }

    def __init__(
        self,
        years: Optional[Union[int, Iterable[int]]] = None,
        expand: bool = True,
        observed: bool = True,
        subdiv: Optional[str] = None,
        prov: Optional[str] = None,
        state: Optional[str] = None,
    ) -> None:
        """
        A subclass of :py:class:`HolidayBase` representing public holidays in
        Singapore.

        Limitations:

        - Prior to 1969: holidays are estimated.
        - Prior to 2000: holidays may not be accurate.
        - 2024 and later: the following four moving date holidays (whose exact
          date is announced yearly) are estimated, and so denoted:

          - Hari Raya Puasa
          - Hari Raya Haji
          - Vesak Day
          - Deepavali

        Sources:

        - `Holidays Act <https://sso.agc.gov.sg/Act/HA1998>`__ (Act 24 of
          1968—Holidays (Amendment) Act 1968)
        - `Ministry of Manpower
          <https://www.mom.gov.sg/employment-practices/public-holidays>`__

        References:

        - `Wikipedia
          <https://en.wikipedia.org/wiki/Public_holidays_in_Singapore>`__

        Country created and maintained by: `Mike Borsetti
        <https://github.com/mborsetti>`__

        See parameters and usage in :py:class:`HolidayBase`.
        """

        self.cnls = _ChineseLuniSolar()
        super().__init__(years, expand, observed, subdiv, prov, state)

    def _populate(self, year) -> None:
        super()._populate(year)

        # New Year's Day
        self[date(year, JAN, 1)] = "New Year's Day"

        # Chinese New Year (two days)
        hol_date = self.cnls.lunar_n_y_date(year)
        self[hol_date] = "Chinese New Year"
        self[hol_date + rd(days=+1)] = "Chinese New Year"

        # Hari Raya Puasa
        # aka Eid al-Fitr
        # Date of observance is announced yearly.
        # An Islamic holiday could fall twice in the same Gregorian year.
        dates_fixed_multiple_obs: Dict[int, Tuple[Tuple[int, int], ...]] = {
            2001: ((DEC, 16),),
            2002: ((DEC, 6),),
            2003: ((NOV, 25),),
            2004: ((NOV, 14),),
            2005: ((NOV, 3),),
            2006: ((OCT, 24),),
            2007: ((OCT, 13),),
            2008: ((OCT, 1),),
            2009: ((SEP, 20),),
            2010: ((SEP, 10),),
            2011: ((AUG, 30),),
            2012: ((AUG, 19),),
            2013: ((AUG, 8),),
            2014: ((JUL, 28),),
            2015: ((JUL, 17),),
            2016: ((JUL, 6),),
            2017: ((JUN, 25),),
            2018: ((JUN, 15),),
            2019: ((JUN, 5),),
            2020: ((MAY, 24),),
            2021: ((MAY, 13),),
            2022: ((MAY, 2),),
            2023: ((APR, 22),),
        }
        if year in dates_fixed_multiple_obs:
            for month_day in dates_fixed_multiple_obs[year]:
                hol_date = date(year, *month_day)
                self[hol_date] = "Hari Raya Puasa"
                # Second day of Hari Raya Puasa (up to and including 1968)
                # Removed since we don't have Hari Raya Puasa dates for the
                # the years <= 1968:
                # if year <= 1968:
                #     self[hol_date + rd(days=+1),
                #                  "Second day of Hari Raya Puasa")
        else:
            for date_obs in _islamic_to_gre(year, 10, 1):
                hol_date = date_obs
                self[hol_date] = "Hari Raya Puasa* (*estimated)"
                # Second day of Hari Raya Puasa (up to and including 1968)
                if year <= 1968:
                    hol_date += rd(days=+1)
                    self[hol_date] = (
                        "Second day of Hari Raya Puasa*" " (*estimated)"
                    )

        # Hari Raya Haji
        # aka Eid al-Adha
        # Date of observance is announced yearly.
        # An Islamic holiday could fall twice in the same Gregorian year.
        dates_fixed_multiple_obs = {
            2001: ((MAR, 6),),
            2002: ((FEB, 23),),
            2003: ((FEB, 12),),
            2004: ((FEB, 1),),
            2005: ((JAN, 21),),
            2006: ((JAN, 10),),
            2007: ((DEC, 20),),
            2008: ((DEC, 8),),
            2009: ((NOV, 27),),
            2010: ((NOV, 17),),
            2011: ((NOV, 6),),
            2012: ((OCT, 26),),
            2013: ((OCT, 15),),
            2014: ((OCT, 5),),
            2015: ((SEP, 24),),
            2016: ((SEP, 12),),
            2017: ((SEP, 1),),
            2018: ((AUG, 22),),
            2019: ((AUG, 11),),
            2020: ((JUL, 31),),
            2021: ((JUL, 20),),
            2022: ((JUL, 9),),
            2023: ((JUN, 29),),
        }
        if year in dates_fixed_multiple_obs:
            for month_day in dates_fixed_multiple_obs[year]:
                hol_date = date(year, *month_day)
                self[hol_date] = "Hari Raya Haji"
        else:
            for date_obs in _islamic_to_gre(year, 12, 10):
                hol_date = date_obs
                self[hol_date] = "Hari Raya Haji* (*estimated)"

        easter_date = easter(year)
        # Good Friday
        self[easter_date + rd(days=-2)] = "Good Friday"

        if year <= 1968:
            # Holy Saturday
            self[easter_date + rd(days=-1)] = "Holy Saturday"

            # Easter Monday
            self[easter_date + rd(days=+1)] = "Easter Monday"

        # Labour Day
        self[date(year, MAY, 1)] = "Labour Day"

        # Vesak Day
        # date of observance is announced yearly
        # https://en.wikipedia.org/wiki/Vesak#Dates_of_observance
        dates_fixed_obs: Dict[int, Tuple[int, int]] = {
            2001: (MAY, 7),
            2002: (MAY, 27),
            2003: (MAY, 15),
            2004: (JUN, 2),
            2005: (MAY, 23),
            2006: (MAY, 12),
            2007: (MAY, 31),
            2008: (MAY, 19),
            2009: (MAY, 9),
            2010: (MAY, 28),
            2011: (MAY, 17),
            2012: (MAY, 5),
            2013: (MAY, 24),
            2014: (MAY, 13),
            2015: (JUN, 1),
            2016: (MAY, 20),
            2017: (MAY, 10),
            2018: (MAY, 29),
            2019: (MAY, 19),
            2020: (MAY, 7),
            2021: (MAY, 26),
            2022: (MAY, 15),
            # 2023 date revised by MOM on 29-sep-22
            # https://www.mom.gov.sg/newsroom/press-releases/2022/0929-revised-date-for-vesak-day-2023
            2023: (JUN, 2),
        }
        if year in dates_fixed_obs:
            hol_date = date(year, *dates_fixed_obs[year])
            self[hol_date] = "Vesak Day"
        else:
            hol_date = self.cnls.vesak_date(year)
            self[hol_date] = "Vesak Day* (*estimated; ~10% chance +/- 1 day)"

        # National Day
        self[date(year, AUG, 9)] = "National Day"

        # Deepavali
        # aka Diwali
        # date of observance is announced yearly
        dates_fixed_obs = {
            2001: (NOV, 14),
            2002: (NOV, 3),
            2003: (OCT, 23),
            2004: (NOV, 11),
            2005: (NOV, 1),
            2006: (OCT, 21),
            2007: (NOV, 8),
            2008: (OCT, 27),
            2009: (OCT, 17),
            2010: (NOV, 5),
            2011: (OCT, 26),
            2012: (NOV, 13),
            2013: (NOV, 2),
            2014: (OCT, 22),
            2015: (NOV, 10),
            2016: (OCT, 29),
            2017: (OCT, 18),
            2018: (NOV, 6),
            2019: (OCT, 27),
            2020: (NOV, 14),
            2021: (NOV, 4),
            2022: (OCT, 24),
            2023: (NOV, 12),
        }
        if year in dates_fixed_obs:
            hol_date = date(year, *dates_fixed_obs[year])
            self[hol_date] = "Deepavali"
        else:
            hol_date = self.cnls.s_diwali_date(year)
            self[hol_date] = "Deepavali* (*estimated; rarely on day after)"

        # Christmas Day
        self[date(year, DEC, 25)] = "Christmas Day"

        # Boxing day (up to and including 1968)
        if year <= 1968:
            self[date(year, DEC, 26)] = "Boxing Day"

        # Check for holidays that fall on a Sunday and implement Section 4(2)
        # of the Holidays Act: "if any day specified in the Schedule falls on
        # a Sunday, the day next following not being itself a public holiday
        # is declared a public holiday in Singapore."
        for (hol_date, hol_name) in list(self.items()):
            if hol_date.year == year and hol_date.weekday() == SUN:
                self[hol_date] += " [Sunday]"
                in_lieu_date = hol_date + rd(days=+1)
                while in_lieu_date in self:
                    in_lieu_date += rd(days=+1)
                self[in_lieu_date] = hol_name + " [In lieu]"


class SG(Singapore):

    # __init__ required for IDE typing and inheritance of docstring.
    def __init__(
        self,
        years: Optional[Union[int, Iterable[int]]] = None,
        expand: bool = True,
        observed: bool = True,
        subdiv: Optional[str] = None,
        prov: Optional[str] = None,
        state: Optional[str] = None,
    ) -> None:
        super().__init__(years, expand, observed, subdiv, prov, state)


class SGP(Singapore):

    # __init__ required for IDE typing and inheritance of docstring.
    def __init__(
        self,
        years: Optional[Union[int, Iterable[int]]] = None,
        expand: bool = True,
        observed: bool = True,
        subdiv: Optional[str] = None,
        prov: Optional[str] = None,
        state: Optional[str] = None,
    ) -> None:
        super().__init__(years, expand, observed, subdiv, prov, state)



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, APR, MAY, JUN, JUL, OCT, DEC
from holidays.holiday_base import HolidayBase


class Venezuela(HolidayBase):
    """
    https://dias-festivos.eu/dias-festivos/venezuela/#
    """

    country = "VE"

    def _populate(self, year):
        super()._populate(year)

        """
        Overview: https://dias-festivos.eu/dias-festivos/venezuela/#
        Various decrees about holidays:
          1909 (AUG 5): https://bit.ly/3J0mWKQ
          1918 (MAY 19): https://bit.ly/3B8O1Jz
          1921 (JUN 11): https://bit.ly/3aUE2gz
          1971 (JUN 22): https://bit.ly/3yZaUN9
          2002 (OCT 10): https://bit.ly/3B7nRqC
          2012 (MAY 7): https://bit.ly/2MT5x97
        """

        self[date(year, JAN, 1)] = "Año Nuevo [New Year's]"

        easter_date = easter(year)
        self[
            easter_date + rd(days=-48)
        ] = "Lunes de Carnaval [Monday of Carnival]"

        self[
            easter_date + rd(days=-47)
        ] = "Martes de Carnaval [Tuesday of Carnival]"

        self[easter_date + rd(days=-3)] = "Jueves Santo [Maundy Thursday]"

        self[easter_date + rd(days=-2)] = "Viernes Santo [Good Friday]"

        # Note: not sure about the start year, but this event happened in 1811
        if year >= 1811:
            self[
                date(year, APR, 19)
            ] = "Declaración de la Independencia [Declaration of Independence]"

        # https://bit.ly/3B4Xd1L
        if year >= 1946:
            self[
                date(year, MAY, 1)
            ] = "Dia Mundial del Trabajador [International Worker's Day]"

        # Note: not sure about the start year, but this event happened in 1824
        if year >= 1971 or (1918 > year >= 1824):
            self[
                date(year, JUN, 24)
            ] = "Batalla de Carabobo [Battle of Carabobo]"

        # Note: not sure about the start year, but this event happened in 1811
        if year >= 1811:
            self[
                date(year, JUL, 5)
            ] = "Día de la Independencia [Independence Day]"

        if year >= 1918:
            self[
                date(year, JUL, 24)
            ] = "Natalicio de Simón Bolívar [Birth of Simon Bolivar]"

        if year >= 2002:
            self[
                date(year, OCT, 12)
            ] = "Día de la Resistencia Indígena [Day of Indigenous Resistance]"
        elif year >= 1921:
            self[date(year, OCT, 12)] = "Día de la Raza [Columbus Day]"

        # Note: not sure about the start year nor the reason this was
        # Note: celebrated; the historical records are unclear
        if 1909 <= year < 1918:
            self[
                date(year, OCT, 28)
            ] = "Día Festivo Desconocido [Unknown Holiday]"

        self[date(year, DEC, 24)] = "Nochebuena [Christmas Eve]"

        self[date(year, DEC, 25)] = "Día de Navidad [Christmas Day]"

        self[date(year, DEC, 31)] = "Fiesta de Fin de Año [New Year's Eve]"


class VE(Venezuela):
    pass


class VEN(Venezuela):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import EASTER_ORTHODOX, easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, APR, MAY, AUG, OCT, NOV
from holidays.holiday_base import HolidayBase


class Georgia(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Georgia_(country)
    """

    country = "GE"

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        name = "ახალი წელი"
        self[date(year, JAN, 1)] = name

        # New Year's Day
        name = "ბედობა"
        self[date(year, JAN, 2)] = name

        # Christmas Day (Orthodox)
        name = "ქრისტეშობა"
        self[date(year, JAN, 7)] = name

        # Baptism Day of our Lord Jesus Christ
        name = "ნათლისღება"
        self[date(year, JAN, 19)] = name

        # Mother's Day
        name = "დედის დღე"
        self[date(year, MAR, 3)] = name

        # Women's Day
        name = "ქალთა საერთაშორისო დღე"
        self[date(year, MAR, 8)] = name

        easter_date = easter(year, method=EASTER_ORTHODOX)
        # Orthodox Good Friday
        name = "წითელი პარასკევი"
        self[easter_date + rd(days=-2)] = name

        # Orthodox Holy Saturday
        name = "დიდი შაბათი"
        self[easter_date + rd(days=-1)] = name

        # 	Orthodox Easter Sunday
        name = "აღდგომა"
        self[easter_date] = name

        # Orthodox Easter Monday
        name = "შავი ორშაბათი"
        self[easter_date + rd(days=+1)] = name

        # National Unity Day
        name = "ეროვნული ერთიანობის დღე"
        self[date(year, APR, 9)] = name

        # Day of Victory
        name = "ფაშიზმზე გამარჯვების დღე"
        self[date(year, MAY, 9)] = name

        # Saint Andrew the First-Called Day
        name = "წმინდა ანდრია პირველწოდებულის დღე"
        self[date(year, MAY, 12)] = name

        # Independence Day
        name = "დამოუკიდებლობის დღე"
        self[date(year, MAY, 26)] = name

        # Saint Mary's Day
        name = "მარიამობა"
        self[date(year, AUG, 28)] = name

        # Day of Svetitskhoveli Cathedral
        name = "სვეტიცხოვლობა"
        self[date(year, OCT, 14)] = name

        # Saint George's Day
        name = "გიორგობა"
        self[date(year, NOV, 23)] = name


class GE(Georgia):
    pass


class GEO(Georgia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT
from holidays.constants import NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class Azerbaijan(HolidayBase):
    # [1] https://en.wikipedia.org/wiki/Public_holidays_in_Azerbaijan
    # [2] https://az.wikipedia.org/wiki/Az%C9%99rbaycan%C4%B1n_d%C3%B6vl%C9%99t_bayramlar%C4%B1_v%C9%99_x%C3%BCsusi_g%C3%BCnl%C9%99ri  # noqa: E501
    # [3] https://www.sosial.gov.az/en/prod-calendar

    country = "AZ"

    def _populate(self, year: int) -> None:
        def _add_observed(hol_date: date, hol_name: str) -> None:
            next_workday = hol_date + rd(days=+1)
            while next_workday.year == year and (
                self._is_weekend(next_workday) or self.get(next_workday)
            ):
                next_workday += rd(days=+1)
            _add_holiday(next_workday, f"{hol_name} (Observed)")

        def _add_holiday(hol_date: date, hol_name: str) -> None:
            if hol_date.year == year:
                self[hol_date] = hol_name

        if year <= 1989:
            return
        super()._populate(year)

        # New Year
        name = "New Year's Day"
        self[date(year, JAN, 1)] = name
        if year >= 2006:
            self[date(year, JAN, 2)] = name

        # Black January
        if year >= 2000:
            self[date(year, JAN, 20)] = "Black January"

        # International Women's Day
        self[date(year, MAR, 8)] = "International Women's Day"

        # Novruz
        if year >= 2007:
            for i in range(20, 25):
                self[date(year, MAR, i)] = "Novruz"

        # Victory Day
        self[date(year, MAY, 9)] = "Victory Day over Fascism"

        # Republic Day
        if year >= 1992:
            name = "Independence Day" if year >= 2021 else "Republic Day"
            self[date(year, MAY, 28)] = name

        # National Salvation Day
        if year >= 1997:
            self[date(year, JUN, 15)] = "National Salvation Day"

        # Memorial Day
        if year >= 2021:
            self[date(year, SEP, 27)] = "Memorial Day"

        # Azerbaijan Armed Forces Day
        if year >= 1992:
            name = "Azerbaijan Armed Forces Day"
            if year <= 1997:
                self[date(year, OCT, 9)] = name
            else:
                self[date(year, JUN, 26)] = name

        # Independence Day
        if year <= 2005:
            self[date(year, OCT, 18)] = "Independence Day"

        # Victory Day
        if year >= 2021:
            self[date(year, NOV, 8)] = "Victory Day"

        # Flag Day
        if year >= 2010:
            self[date(year, NOV, 9)] = "Flag Day"

        # International Solidarity Day of Azerbaijanis
        if year >= 1993:
            name = "International Solidarity Day of Azerbaijanis"
            self[date(year, DEC, 31)] = name
            if self.observed and year >= 2006:
                dt = date(year - 1, DEC, 31)
                if self._is_weekend(dt):
                    _add_observed(dt, name)

        ramazan_dates_obs = {
            2011: ((AUG, 30),),
            2012: ((AUG, 19),),
            2013: ((AUG, 8),),
            2014: ((JUL, 28),),
            2015: ((JUL, 17),),
            2016: ((JUL, 6),),
            2017: ((JUN, 26),),
            2018: ((JUN, 15),),
            2019: ((JUN, 5),),
            2020: ((MAY, 24),),
            2021: ((MAY, 13),),
            2022: ((MAY, 2),),
            2023: ((APR, 21),),
        }

        gurban_dates_obs = {
            2011: ((NOV, 6),),
            2012: ((OCT, 25),),
            2013: ((OCT, 15),),
            2014: ((OCT, 4),),
            2015: ((SEP, 24),),
            2016: ((SEP, 12),),
            2017: ((SEP, 1),),
            2018: ((AUG, 22),),
            2019: ((AUG, 12),),
            2020: ((JUL, 31),),
            2021: ((JUL, 20),),
            2022: ((JUL, 9),),
            2023: ((JUN, 28),),
        }

        religious_holidays = (
            ("Ramazan Bayrami", 10, 1, ramazan_dates_obs),
            ("Gurban Bayrami", 12, 10, gurban_dates_obs),
        )
        if year >= 1993:
            for name, hmonth, hday, dates_obs in religious_holidays:
                for yr in (year - 1, year):
                    if yr in dates_obs:
                        for date_obs in dates_obs[yr]:
                            _add_holiday(date(yr, *date_obs), name)
                            _add_holiday(
                                date(yr, *date_obs) + rd(days=+1), name
                            )
                    else:
                        for dt in _islamic_to_gre(yr, hmonth, hday):
                            _add_holiday(dt, f"{name}* (*estimated)")
                            _add_holiday(
                                dt + rd(days=+1), f"{name}* (*estimated)"
                            )

        """
        [2]:
        Article 105 of the Labor Code of the Republic of Azerbaijan states:
        5. If interweekly rest days and holidays that are not considered
        working days overlap, that rest day is immediately transferred to
        the next working day.
        6. If the holidays of Qurban and Ramadan coincide with another holiday
        that is not considered a working day, the next working day is
         considered a rest day. *(machine translated)*
        """
        if self.observed and year >= 2006:
            for k, v in list(self.items()):
                if (
                    k.year == year
                    and self._is_weekend(k)
                    and k
                    not in (
                        date(year, JAN, 20),
                        date(year, SEP, 27),
                    )
                ):
                    _add_observed(k, v)
                hol_names = self.get_list(k)
                if len(hol_names) > 1 and " (Observed)" not in v:
                    for name in hol_names:
                        if name in {hol[0] for hol in religious_holidays}:
                            _add_observed(k, name)


class AZ(Azerbaijan):
    pass


class AZE(Azerbaijan):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from typing import Dict, Tuple, List

from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP
from holidays.constants import OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.utils import _islamic_to_gre


class Pakistan(HolidayBase):
    country = "PK"

    def _populate(self, year):
        def _add_holiday(dt: date, hol: str) -> None:
            if dt.year == year:
                self[dt] = hol

        if year <= 1947:
            return
        super()._populate(year)

        # Kashmir Solidarity Day
        if year >= 1990:
            self[date(year, FEB, 5)] = "Kashmir Solidarity Day"

        # Pakistan Day
        if year >= 1956:
            self[date(year, MAR, 23)] = "Pakistan Day"

        # Labour Day
        if year >= 1972:
            self[date(year, MAY, 1)] = "Labour Day"

        # Independence Day
        self[date(year, AUG, 14)] = "Independence Day"

        # Iqbal Day
        if year <= 2014 or year >= 2022:
            self[date(year, NOV, 9)] = "Iqbal Day"

        # Quaid-e-Azam Day
        self[date(year, DEC, 25)] = "Quaid-e-Azam Day"

        # Eid-ul-Fitr
        # https://www.timeanddate.com/holidays/pakistan/eid-ul-fitr-1
        dates_obs = {
            2005: ((NOV, 4),),
            2006: ((OCT, 24),),
            2007: ((OCT, 13),),
            2008: ((OCT, 2),),
            2009: ((SEP, 21),),
            2010: ((SEP, 10),),
            2011: ((AUG, 31),),
            2012: ((AUG, 19),),
            2013: ((AUG, 8),),
            2014: ((JUL, 29),),
            2015: ((JUL, 17),),
            2016: ((JUL, 6),),
            2017: ((JUN, 26),),
            2018: ((JUN, 16),),
            2019: ((JUN, 5),),
            2020: ((MAY, 24),),
            2021: ((MAY, 13),),
            2022: ((MAY, 3),),
        }
        name = "Eid-ul-Fitr"
        hol_dates = self._get_islamic_holiday(name, year, 10, 1, dates_obs)
        for hol_date, hol_name in hol_dates:
            _add_holiday(hol_date, hol_name)
            _add_holiday(hol_date + rd(days=+1), hol_name)
            _add_holiday(hol_date + rd(days=+2), hol_name)

        # Eid-ul-Adha
        # https://www.timeanddate.com/holidays/pakistan/eid-ul-azha
        dates_obs = {
            2005: ((JAN, 21),),
            2006: (
                (JAN, 10),
                (DEC, 31),
            ),
            2007: ((DEC, 20),),
            2008: ((DEC, 9),),
            2009: ((NOV, 28),),
            2010: ((NOV, 17),),
            2011: ((NOV, 7),),
            2012: ((OCT, 26),),
            2013: ((OCT, 15),),
            2014: ((OCT, 6),),
            2015: ((SEP, 24),),
            2016: ((SEP, 12),),
            2017: ((SEP, 2),),
            2018: ((AUG, 22),),
            2019: ((AUG, 12),),
            2020: ((JUL, 31),),
            2021: ((JUL, 21),),
            2022: ((JUL, 10),),
        }
        name = "Eid-ul-Adha"
        hol_dates = self._get_islamic_holiday(name, year, 12, 10, dates_obs)
        for hol_date, hol_name in hol_dates:
            _add_holiday(hol_date, hol_name)
            _add_holiday(hol_date + rd(days=+1), hol_name)
            _add_holiday(hol_date + rd(days=+2), hol_name)

        # Eid Milad-un-Nabi, Birth of the Prophet
        # https://www.timeanddate.com/holidays/pakistan/eid-milad-un-nabi
        dates_obs = {
            2005: ((APR, 22),),
            2006: ((APR, 11),),
            2007: ((MAR, 31),),
            2008: ((MAR, 21),),
            2009: ((MAR, 9),),
            2010: ((MAR, 1),),
            2011: ((FEB, 17),),
            2012: ((FEB, 5),),
            2013: ((JAN, 24),),
            2014: ((JAN, 14),),
            2015: ((JAN, 4),),
            2016: ((DEC, 12),),
            2017: ((DEC, 1),),
            2018: ((NOV, 21),),
            2019: ((NOV, 10),),
            2020: ((OCT, 30),),
            2021: ((OCT, 19),),
            2022: ((OCT, 9),),
        }
        name = "Eid Milad-un-Nabi"
        hol_dates = self._get_islamic_holiday(name, year, 3, 12, dates_obs)
        for hol_date, hol_name in hol_dates:
            _add_holiday(hol_date, hol_name)

        # Ashura
        # https://www.timeanddate.com/holidays/pakistan/first-day-ashura
        dates_obs = {
            2005: ((FEB, 18),),
            2006: ((FEB, 8),),
            2007: ((JAN, 28),),
            2008: ((JAN, 18),),
            2009: (
                (JAN, 6),
                (DEC, 26),
            ),
            2010: ((DEC, 16),),
            2011: ((DEC, 5),),
            2012: ((NOV, 23),),
            2013: ((NOV, 13),),
            2014: ((NOV, 3),),
            2015: ((OCT, 23),),
            2016: ((OCT, 11),),
            2017: ((SEP, 30),),
            2018: ((SEP, 21),),
            2019: ((SEP, 9),),
            2020: ((AUG, 29),),
            2021: ((AUG, 18),),
            2022: ((AUG, 9),),
        }
        name = "Ashura"
        hol_dates = self._get_islamic_holiday(name, year, 1, 10, dates_obs)
        for hol_date, hol_name in hol_dates:
            _add_holiday(hol_date, hol_name)
            _add_holiday(hol_date + rd(days=+1), hol_name)

    @staticmethod
    def _get_islamic_holiday(
        name: str,
        year: int,
        hmonth: int,
        hday: int,
        known_dates: Dict[int, Tuple[Tuple[int, int]]],
    ) -> List[Tuple[date, str]]:
        """
         Return the Gregorian dates of all instances of Islamic holiday
         falling within specified and previous Gregorian year
         (used to multi-day Islamic holidays).
         If specified year is in known dates list, known date used,
         else calculated date.

        :param name:
            Holiday name

        :param year:
            Gregorian year

        :param hmonth:
            Hijri month of holiday

        :param hday:
            Hijri day of holiday

        :param known_dates:
            Known holiday dates list

        :return:
            List of date-name pairs
        """
        hol_dates: List[Tuple[date, str]] = []
        for yr in (year - 1, year):
            if yr in known_dates:
                for date_obs in known_dates[yr]:
                    hol_dates.append((date(yr, *date_obs), name))
            else:
                for dt in _islamic_to_gre(yr, hmonth, hday):
                    hol_dates.append((dt, f"{name}* (*estimated)"))
        return hol_dates


class PK(Pakistan):
    pass


class PAK(Pakistan):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import MO, SU
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP
from holidays.constants import OCT, NOV, DEC, FRI, SUN
from holidays.holiday_base import HolidayBase


class Canada(HolidayBase):
    country = "CA"
    subdivisions = [
        "AB",
        "BC",
        "MB",
        "NB",
        "NL",
        "NS",
        "NT",
        "NU",
        "ON",
        "PE",
        "QC",
        "SK",
        "YT",
    ]

    def __init__(self, **kwargs):
        # Default subdivision to ON; prov for backwards compatibility
        if not kwargs.get("subdiv", kwargs.get("prov")):
            kwargs["subdiv"] = "ON"
        HolidayBase.__init__(self, **kwargs)

    @staticmethod
    def _get_nearest_monday(d: date) -> date:
        if d.weekday() < FRI:
            return d + rd(weekday=MO(-1))
        else:
            return d + rd(weekday=MO)

    def _populate(self, year):
        super()._populate(year)

        if year < 1867:
            return

        # New Year's Day
        name = "New Year's Day"
        self[date(year, JAN, 1)] = name
        if self.observed and self._is_weekend(date(year, JAN, 1)):
            self[date(year, JAN, 1) + rd(weekday=MO)] = name + " (Observed)"

        # Family Day / Louis Riel Day (MB) / Islander Day (PE)
        # / Heritage Day (NS, YT)
        if (
            (self.subdiv == "AB" and year >= 1990)
            or (self.subdiv == "SK" and year >= 2007)
            or (self.subdiv == "ON" and year >= 2008)
            or (self.subdiv == "NB" and year >= 2018)
        ):
            self[date(year, FEB, 1) + rd(weekday=MO(+3))] = "Family Day"
        elif self.subdiv == "BC":
            if 2013 <= year <= 2018:
                self[date(year, FEB, 1) + rd(weekday=MO(+2))] = "Family Day"
            elif year > 2018:
                self[date(year, FEB, 1) + rd(weekday=MO(+3))] = "Family Day"
        elif self.subdiv == "MB" and year >= 2008:
            self[date(year, FEB, 1) + rd(weekday=MO(+3))] = "Louis Riel Day"
        elif self.subdiv == "PE" and year >= 2010:
            self[date(year, FEB, 1) + rd(weekday=MO(+3))] = "Islander Day"
        elif self.subdiv == "PE" and year == 2009:
            self[date(year, FEB, 1) + rd(weekday=MO(+2))] = "Islander Day"
        elif self.subdiv == "NS" and year >= 2015:
            # http://novascotia.ca/lae/employmentrights/NovaScotiaHeritageDay.asp
            self[date(year, FEB, 1) + rd(weekday=MO(+3))] = "Heritage Day"
        elif self.subdiv == "YT" and year >= 1974:
            # start date?
            # https://www.britannica.com/topic/Heritage-Day-Canadian-holiday
            # Heritage Day was created in 1973
            # by the Heritage Canada Foundation
            # therefore, start date is not earlier than 1974
            # http://heritageyukon.ca/programs/heritage-day
            # https://en.wikipedia.org/wiki/Family_Day_(Canada)#Yukon_Heritage_Day
            # Friday before the last Sunday in February
            dt = (
                date(year, MAR, 1)
                + rd(days=-1)
                + rd(weekday=SU(-1))
                + rd(days=-2)
            )
            self[dt] = "Heritage Day"

        # St. Patrick's Day
        if self.subdiv == "NL" and year >= 1900:
            # Nearest Monday to March 17
            dt = self._get_nearest_monday(date(year, MAR, 17))
            self[dt] = "St. Patrick's Day"

        easter_date = easter(year)
        # Good Friday
        self[easter_date + rd(days=-2)] = "Good Friday"
        # Easter Monday
        self[easter_date + rd(days=+1)] = "Easter Monday"

        # St. George's Day
        if self.subdiv == "NL" and year >= 1990:
            if year == 2010:
                # 4/26 is the Monday closer to 4/23 in 2010
                # but the holiday was observed on 4/19? Crazy Newfies!
                dt = date(2010, APR, 19)
            else:
                # Nearest Monday to April 23
                dt = self._get_nearest_monday(date(year, APR, 23))
            self[dt] = "St. George's Day"

        # Victoria Day / National Patriots' Day (QC)
        if year >= 1953:
            dt = date(year, MAY, 24) + rd(weekday=MO(-1))
            if self.subdiv not in {"NB", "NS", "PE", "NL", "QC"}:
                self[dt] = "Victoria Day"
            elif self.subdiv == "QC":
                self[dt] = "National Patriots' Day"

        # National Aboriginal Day
        if self.subdiv == "NT" and year >= 1996:
            self[date(year, JUN, 21)] = "National Aboriginal Day"

        # St. Jean Baptiste Day
        if self.subdiv == "QC" and year >= 1925:
            name = "St. Jean Baptiste Day"
            dt = date(year, JUN, 24)
            self[dt] = name
            if self.observed and dt.weekday() == SUN:
                self[dt + rd(days=1)] = name + " (Observed)"

        # Discovery Day
        if self.subdiv == "NL" and year >= 1997:
            # Nearest Monday to June 24
            dt = self._get_nearest_monday(date(year, JUN, 24))
            self[dt] = "Discovery Day"
        elif self.subdiv == "YT" and year >= 1912:
            self[date(year, AUG, 1) + rd(weekday=MO(+3))] = "Discovery Day"

        # Canada Day / Memorial Day (NL)
        if year >= 1983:
            name = "Memorial Day" if self.subdiv == "NL" else "Canada Day"
        else:
            name = "Dominion Day"
        dt = date(year, JUL, 1)
        self[dt] = name
        if year >= 1879 and self.observed and self._is_weekend(dt):
            self[dt + rd(weekday=MO)] = name + " (Observed)"

        # Nunavut Day
        if self.subdiv == "NU":
            name = "Nunavut Day"
            if year >= 2001:
                dt = date(year, JUL, 9)
                self[dt] = name
                if self.observed and dt.weekday() == SUN:
                    self[dt + rd(days=1)] = name + " (Observed)"
            elif year == 2000:
                self[date(2000, APR, 1)] = name

        # Civic Holiday
        if year >= 1900 and self.subdiv in {"MB", "NT", "ON"}:
            self[date(year, AUG, 1) + rd(weekday=MO)] = "Civic Holiday"
        elif year >= 1974 and self.subdiv == "AB":
            # https://en.wikipedia.org/wiki/Civic_Holiday#Alberta
            self[date(year, AUG, 1) + rd(weekday=MO)] = "Heritage Day"
        elif year >= 1974 and self.subdiv == "BC":
            # https://en.wikipedia.org/wiki/Civic_Holiday
            self[date(year, AUG, 1) + rd(weekday=MO)] = "British Columbia Day"
        elif year >= 1900 and self.subdiv == "NB":
            # https://en.wikipedia.org/wiki/Civic_Holiday
            self[date(year, AUG, 1) + rd(weekday=MO)] = "New Brunswick Day"
        elif year >= 1900 and self.subdiv == "SK":
            # https://en.wikipedia.org/wiki/Civic_Holiday
            self[date(year, AUG, 1) + rd(weekday=MO)] = "Saskatchewan Day"

        # Labour Day
        if year >= 1894:
            self[date(year, SEP, 1) + rd(weekday=MO)] = "Labour Day"

        # Funeral of Queen Elizabeth II
        # https://www.narcity.com/provinces-territories-will-have-a-day-off-monday-mourn-queen
        # TODO: the territories holiday status (NT, NU, YT) is still tentative
        if year == 2022 and self.subdiv in {
            "BC",
            "NB",
            "NL",
            "NS",
            "PE",
            "YT",
        }:
            self[
                date(2022, SEP, 19)
            ] = "Funeral of Her Majesty the Queen Elizabeth II"

        # National Day for Truth and Reconciliation
        if year >= 2021 and self.subdiv in {"MB", "NS"}:
            self[
                date(year, SEP, 30)
            ] = "National Day for Truth and Reconciliation"

        # Thanksgiving
        if year >= 1931 and self.subdiv not in {"NB", "NL", "NS", "PE"}:
            if year == 1935:
                # in 1935, Canadian Thanksgiving was moved due to the General
                # Election falling on the second Monday of October
                # https://books.google.ca/books?id=KcwlQsmheG4C&pg=RA1-PA1940&lpg=RA1-PA1940&dq=canada+thanksgiving+1935&source=bl&ots=j4qYrcfGuY&sig=gxXeAQfXVsOF9fOwjSMswPHJPpM&hl=en&sa=X&ved=0ahUKEwjO0f3J2PjOAhVS4mMKHRzKBLAQ6AEIRDAG#v=onepage&q=canada%20thanksgiving%201935&f=false
                self[date(1935, OCT, 25)] = "Thanksgiving"
            else:
                self[date(year, OCT, 1) + rd(weekday=MO(+2))] = "Thanksgiving"

        # Remembrance Day
        if year >= 1931 and self.subdiv not in {"ON", "QC"}:
            name = "Remembrance Day"
            dt = date(year, NOV, 11)
            self[dt] = name
            if (
                self.observed
                and dt.weekday() == SUN
                and self.subdiv in {"NS", "NL", "NT", "PE", "SK"}
            ):
                self[dt + rd(weekday=MO)] = name + " (Observed)"

        # Christmas Day
        name = "Christmas Day"
        dt = date(year, DEC, 25)
        self[dt] = name
        if self.observed and self._is_weekend(dt):
            self[dt + rd(days=2)] = name + " (Observed)"

        # Boxing Day
        name = "Boxing Day"
        dt = date(year, DEC, 26)
        self[dt] = name
        if self.observed and self._is_weekend(dt):
            self[dt + rd(days=2)] = name + " (Observed)"


class CA(Canada):
    pass


class CAN(Canada):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import EASTER_ORTHODOX, easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAR, MAY, JUN, AUG, OCT, DEC
from holidays.holiday_base import HolidayBase


class Moldova(HolidayBase):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Moldova
    """

    country = "MD"

    def _populate(self, year):
        super()._populate(year)

        easter_date = easter(year, method=EASTER_ORTHODOX)

        # New Year
        self[date(year, JAN, 1)] = "Anul Nou"

        # Orthodox Christmas
        for day in [7, 8]:
            self[date(year, JAN, day)] = "Crăciunul"

        # International Women's Day
        self[date(year, MAR, 8)] = "Ziua Internatională a Femeii"

        # Orthodox Easter
        for day_after_easter in [-2, 0, 1]:
            self[easter_date + rd(days=day_after_easter)] = "Paştele"

        # Paştele Blajinilor
        self[easter_date + rd(days=+9)] = "Paştele Blajinilor"

        # Labour Day
        self[date(year, MAY, 1)] = "Ziua Internatională a Muncii"

        # Ziua Victoriei
        self[date(year, MAY, 9)] = "Ziua Victoriei"

        # International Children's Day
        self[date(year, JUN, 1)] = "Ziua Copilului"

        # Ziua Independenţei
        self[date(year, AUG, 27)] = "Ziua Independenţei"

        # Limba noastră
        self[date(year, AUG, 31)] = "Limba noastră"

        # Ziua Natională a Vinului
        self[date(year, OCT, 8)] = "Ziua Natională a Vinului"

        # Crăciunul
        self[date(year, DEC, 25)] = "Crăciunul"


class MD(Moldova):
    pass


class MDA(Moldova):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, APR, MAY, JUL, AUG, OCT, DEC, SUN
from holidays.holiday_base import HolidayBase


class Curacao(HolidayBase):
    """
    https://loketdigital.gobiernu.cw/Loket/product/571960bbe1e5fe8712b10a1323630e70
    """

    country = "CW"

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self[date(year, JAN, 1)] = "Nieuwjaarsdag [New Year's Day]"

        easter_date = easter(year)
        # Carnaval Monday
        self[
            easter_date + rd(days=-48)
        ] = "Maandag na de Grote Karnaval [Carnaval Monday]"

        # Good Friday
        self[easter_date + rd(days=-2)] = "Goede Vrijdag [Good Friday]"

        # Easter Monday
        self[
            easter_date + rd(days=+1)
        ] = "Di Dos Dia di Pasku di Resureccion [Easter Monday]"

        # King's Day
        if year >= 2014:
            kings_day = date(year, APR, 27)
            if kings_day.weekday() == SUN:
                kings_day = kings_day - rd(days=1)

            self[kings_day] = "Koningsdag [King's Day]"

        # Queen's Day
        if 1891 <= year <= 2013:
            queens_day = date(year, APR, 30)
            if year <= 1948:
                queens_day = date(year, AUG, 31)

            if queens_day.weekday() == SUN:
                if year < 1980:
                    queens_day = queens_day + rd(days=1)
                else:
                    queens_day = queens_day - rd(days=1)

            self[queens_day] = "Anja di La Reina [Queen's Day]"

        # Labour Day
        labour_day = date(year, MAY, 1)
        if labour_day.weekday() == SUN:
            labour_day = labour_day + rd(days=1)
        self[labour_day] = "Dia di Obrero [Labour Day]"

        # Ascension Day
        self[easter_date + rd(days=+39)] = "Hemelvaartsdag [Ascension Day]"

        # Dia di Himno y Bandera
        self[
            date(year, JUL, 2)
        ] = "Dia di Himno y Bandera [National Anthem & Flag Day]"

        # Dia di Pais Kòrsou
        self[date(year, OCT, 10)] = "Dia di Pais Kòrsou [Curaçao Day]"

        # Christmas Day
        self[date(year, DEC, 25)] = "Kerstdag [Christmas]"

        # Second Christmas
        self[date(year, DEC, 26)] = "2de Kerstdag [Second Christmas]"


class CW(Curacao):
    pass


class CUW(Curacao):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from dateutil.easter import easter
from dateutil.relativedelta import relativedelta as rd

from holidays.constants import JAN, MAY, AUG, NOV, DEC
from holidays.holiday_base import HolidayBase


class Poland(HolidayBase):
    """
    https://pl.wikipedia.org/wiki/Dni_wolne_od_pracy_w_Polsce
    """

    country = "PL"
    special_holidays = {
        2018: ((NOV, 12, "Narodowe Święto Niepodległości - 100-lecie"),)
    }

    def _populate(self, year):
        super()._populate(year)

        self[date(year, JAN, 1)] = "Nowy Rok"
        if year >= 2011:
            self[date(year, JAN, 6)] = "Święto Trzech Króli"

        easter_date = easter(year)
        self[easter_date] = "Niedziela Wielkanocna"
        self[easter_date + rd(days=+1)] = "Poniedziałek Wielkanocny"

        if year >= 1950:
            self[date(year, MAY, 1)] = "Święto Państwowe"
        if year >= 1919:
            self[date(year, MAY, 3)] = "Święto Narodowe Trzeciego Maja"

        self[easter_date + rd(days=+49)] = "Zielone Świątki"
        self[easter_date + rd(days=+60)] = "Dzień Bożego Ciała"

        self[date(year, AUG, 15)] = "Wniebowzięcie Najświętszej Marii Panny"

        self[date(year, NOV, 1)] = "Uroczystość Wszystkich Świętych"
        if (1937 <= year <= 1945) or year >= 1989:
            self[date(year, NOV, 11)] = "Narodowe Święto Niepodległości"

        self[date(year, DEC, 25)] = "Boże Narodzenie (pierwszy dzień)"
        self[date(year, DEC, 26)] = "Boże Narodzenie (drugi dzień)"


class PL(Poland):
    pass


class POL(Poland):
    pass
